@@ -18,14 +18,13 @@ use tcgeneric::{TCBoxStream, TCBoxTryFuture, TCBoxTryStream};
 use crate::sparse::{SparseAccess, SparseAccessor};
 use crate::stream::{Read, ReadValueAt};
 use crate::{
-    transform, AxisBounds, Bounds, Coord, Phantom, Shape, TensorAccess, TensorType, ERR_INF,
+    transform, Bounds, Coord, Phantom, Shape, TensorAccess, TensorReduce, TensorType, ERR_INF,
     ERR_NAN,
 };
 
 use super::file::{BlockListFile, BlockListFileSlice};
 use super::stream::SparseValueStream;
 use super::{DenseTensor, PER_BLOCK};
-use std::fmt::{Formatter, Write};
 
 /// Common [`DenseTensor`] access methods
 #[async_trait]
@@ -615,6 +614,7 @@ where
             let left = self.left.read_value_at(txn.clone(), coord.to_vec());
             let right = self.right.read_value_at(txn, coord);
             let ((coord, left), (_, right)) = try_join!(left, right)?;
+
             let value = (self.value_combinator)(left, right);
             if value.is_infinite() {
                 Err(TCError::unsupported(ERR_INF))
@@ -670,7 +670,8 @@ where
     B: DenseAccess<FD, FS, D, T>,
 {
     fn dtype(&self) -> NumberType {
-        Ord::max(self.source.dtype(), self.other.class())
+        let combinator = self.value_combinator;
+        combinator(self.source.dtype().zero(), self.other.class().zero()).class()
     }
 
     fn ndim(&self) -> usize {
@@ -712,7 +713,7 @@ where
 
     fn block_stream<'a>(self, txn: T) -> TCBoxTryFuture<'a, TCBoxTryStream<'a, Array>> {
         Box::pin(async move {
-            let combinator = (self.combinator);
+            let combinator = self.combinator;
             let right = self.other;
 
             let left = self.source.block_stream(txn).await?;
@@ -881,24 +882,14 @@ where
     }
 
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
-        debug!("slice {} from broadcast of {}", bounds, self.source.shape());
         self.shape().validate_bounds(&bounds)?;
-
         let shape = bounds.to_shape(self.shape())?;
         let bounds = self.rebase.invert_bounds(bounds);
-        debug!("source bounds are {}, broadcast into {}", bounds, shape);
-
         let source = self.source.slice(bounds)?;
         BlockListBroadcast::new(source, shape)
     }
 
     fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
-        debug!(
-            "BlockListTranspose {:?} (shape is {})",
-            permutation,
-            self.shape()
-        );
-
         BlockListTranspose::new(self, permutation)
     }
 
@@ -970,7 +961,7 @@ where
     B: DenseAccess<FD, FS, D, T>,
 {
     fn dtype(&self) -> NumberType {
-        self.source.dtype()
+        self.dtype
     }
 
     fn ndim(&self) -> usize {
@@ -1125,8 +1116,8 @@ where
     D::FileClass: From<TensorType>,
     B: DenseAccess<FD, FS, D, T>,
 {
-    type Slice = B::Slice;
-    type Transpose = B::Transpose;
+    type Slice = DenseAccessor<FD, FS, D, T>;
+    type Transpose = BlockListExpand<FD, FS, D, T, B::Transpose>;
 
     fn accessor(self) -> DenseAccessor<FD, FS, D, T> {
         let expand = BlockListExpand {
@@ -1149,13 +1140,43 @@ where
     fn slice(self, mut bounds: Bounds) -> TCResult<Self::Slice> {
         self.shape().validate_bounds(&bounds)?;
         bounds.normalize(self.shape());
+        let ndim = bounds.ndim();
+
+        let expand_axis = self.rebase.invert_axis(&bounds);
         let bounds = self.rebase.invert_bounds(bounds);
-        self.source.slice(bounds) // TODO: expand the result
+        let source = self.source.slice(bounds)?;
+
+        if ndim == source.ndim() {
+            Ok(source.accessor())
+        } else if let Some(axis) = expand_axis {
+            let rebase = transform::Expand::new(source.shape().clone(), axis)?;
+            let slice = BlockListExpand {
+                source,
+                rebase,
+                phantom: self.phantom,
+            };
+
+            Ok(slice.accessor())
+        } else {
+            Ok(source.accessor())
+        }
     }
 
     fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
+        let expand_axis = if let Some(permutation) = &permutation {
+            permutation[self.rebase.expand_axis()]
+        } else {
+            self.ndim() - self.rebase.expand_axis()
+        };
+
         let permutation = permutation.map(|axes| self.rebase.invert_axes(axes));
-        self.source.transpose(permutation) // TODO: expand the result
+        let source = self.source.transpose(permutation)?;
+        let rebase = transform::Expand::new(source.shape().clone(), expand_axis)?;
+        Ok(BlockListExpand {
+            source,
+            rebase,
+            phantom: self.phantom,
+        })
     }
 
     async fn read_values(self, txn: Self::Txn, coords: Coords) -> TCResult<Array> {
@@ -1193,15 +1214,41 @@ impl<FD, FS, D, T, B> fmt::Display for BlockListExpand<FD, FS, D, T, B> {
     }
 }
 
-// TODO: &Txn, not Txn
-type Reductor<FD, FS, D, T> =
+#[derive(Copy, Clone)]
+pub enum Reductor {
+    Product(NumberType, u64),
+    Sum(NumberType, u64),
+}
+
+impl Reductor {
+    fn dtype(&self) -> NumberType {
+        match self {
+            Self::Product(dtype, _) => *dtype,
+            Self::Sum(dtype, _) => *dtype,
+        }
+    }
+
+    fn call(self, blocks: TCBoxTryStream<Array>) -> TCBoxTryStream<Array> {
+        let reduced = match self {
+            Self::Product(dtype, stride) => {
+                afarray::reduce_product(blocks, dtype, PER_BLOCK, stride)
+            }
+            Self::Sum(dtype, stride) => afarray::reduce_sum(blocks, dtype, PER_BLOCK, stride),
+        };
+
+        std::pin::Pin::new(reduced)
+    }
+}
+
+type ReduceAll<FD, FS, D, T> =
     fn(&DenseTensor<FD, FS, D, T, DenseAccessor<FD, FS, D, T>>, T) -> TCBoxTryFuture<Number>;
 
 #[derive(Clone)]
 pub struct BlockListReduce<FD, FS, D, T, B> {
     source: B,
     rebase: transform::Reduce,
-    reductor: Reductor<FD, FS, D, T>,
+    reductor: Reductor,
+    reduce_all: ReduceAll<FD, FS, D, T>,
 }
 
 impl<FD, FS, D, T, B> BlockListReduce<FD, FS, D, T, B>
@@ -1209,16 +1256,34 @@ where
     FD: File<Array>,
     FS: File<Node>,
     D: Dir,
+    D::File: AsType<FD> + AsType<FS>,
+    D::FileClass: From<TensorType>,
     T: Transaction<D>,
     B: DenseAccess<FD, FS, D, T>,
 {
-    pub fn new(source: B, axis: usize, reductor: Reductor<FD, FS, D, T>) -> TCResult<Self> {
+    pub fn product(source: B, axis: usize) -> TCResult<Self> {
         let rebase = transform::Reduce::new(source.shape().clone(), axis)?;
+        let dtype = afarray::product_dtype(source.dtype());
+        let stride = source.size() / (source.size() / source.shape()[axis]);
 
         Ok(BlockListReduce {
             source,
             rebase,
-            reductor,
+            reductor: Reductor::Product(dtype, stride),
+            reduce_all: TensorReduce::product_all,
+        })
+    }
+
+    pub fn sum(source: B, axis: usize) -> TCResult<Self> {
+        let rebase = transform::Reduce::new(source.shape().clone(), axis)?;
+        let dtype = afarray::sum_dtype(source.dtype());
+        let stride = source.size() / (source.size() / source.shape()[axis]);
+
+        Ok(BlockListReduce {
+            source,
+            rebase,
+            reductor: Reductor::Sum(dtype, stride),
+            reduce_all: TensorReduce::sum_all,
         })
     }
 }
@@ -1232,7 +1297,7 @@ where
     B: DenseAccess<FD, FS, D, T>,
 {
     fn dtype(&self) -> NumberType {
-        self.source.dtype()
+        self.reductor.dtype()
     }
 
     fn ndim(&self) -> usize {
@@ -1267,38 +1332,45 @@ where
             source: self.source.accessor(),
             rebase: self.rebase,
             reductor: self.reductor,
+            reduce_all: self.reduce_all,
         };
 
         DenseAccessor::Reduce(Box::new(reduce))
     }
 
-    fn value_stream<'a>(self, txn: T) -> TCBoxTryFuture<'a, TCBoxTryStream<'a, Number>> {
+    fn block_stream<'a>(self, txn: Self::Txn) -> TCBoxTryFuture<'a, TCBoxTryStream<'a, Array>> {
         Box::pin(async move {
-            let values = stream::iter(Bounds::all(self.shape()).affected())
-                .map(move |coord| {
-                    let txn = txn.clone();
-                    let source = self.source.clone();
-                    let reductor = self.reductor;
-                    let source_bounds = self.rebase.invert_coord(&coord);
-                    Box::pin(async move {
-                        debug!("slice {} from {}", source_bounds, source);
-                        let slice = source.slice(source_bounds)?;
-                        reductor(&slice.accessor().into(), txn.clone()).await
-                    })
-                })
-                .buffered(num_cpus::get());
+            let reductor = self.reductor;
+            let axis = self.rebase.reduce_axis();
+            let ndim = self.source.ndim();
+            let source = self.source;
 
-            let values: TCBoxTryStream<'a, Number> = Box::pin(values);
-            Ok(values)
+            if axis == ndim - 1 {
+                let blocks = source.block_stream(txn).await?;
+                Ok(reductor.call(blocks))
+            } else {
+                let mut permutation: Vec<usize> = (0..ndim).collect();
+                permutation[axis] = ndim - 1;
+                permutation[ndim - 1] = axis;
+
+                let transpose = source.transpose(Some(permutation))?;
+                let blocks = transpose.block_stream(txn).await?;
+                Ok(reductor.call(blocks))
+            }
         })
     }
 
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
         self.shape().validate_bounds(&bounds)?;
-        let reduce_axis = self.rebase.reduce_axis(&bounds);
+        let reductor = self.reductor;
+        let reduce_axis = self.rebase.invert_axis(&bounds);
         let source_bounds = self.rebase.invert_bounds(bounds);
         let slice = self.source.slice(source_bounds)?;
-        BlockListReduce::new(slice, reduce_axis, self.reductor)
+
+        match reductor {
+            Reductor::Product(_, _) => BlockListReduce::product(slice, reduce_axis),
+            Reductor::Sum(_, _) => BlockListReduce::sum(slice, reduce_axis),
+        }
     }
 
     fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
@@ -1339,11 +1411,10 @@ where
     fn read_value_at<'a>(self, txn: Self::Txn, coord: Coord) -> Read<'a> {
         Box::pin(async move {
             self.shape().validate_coord(&coord)?;
-            let reductor = self.reductor;
+            let reductor = self.reduce_all;
             let source_bounds = self.rebase.invert_coord(&coord);
             let slice = self.source.slice(source_bounds)?;
-            let value = reductor(&slice.accessor().into(), txn.clone()).await?;
-
+            let value = reductor(&slice.accessor().into(), txn).await?;
             Ok((coord, value))
         })
     }
