@@ -5,7 +5,7 @@ use std::ops::{Add, Div, Mul, Sub};
 use afarray::{Array, ArrayInstance};
 use arrayfire as af;
 use async_trait::async_trait;
-use destream::{de, en, EncodeSeq};
+use destream::{de, en};
 use futures::future::{self, TryFutureExt};
 use futures::stream::{Stream, StreamExt, TryStreamExt};
 use log::{debug, warn};
@@ -16,26 +16,26 @@ use tc_error::*;
 use tc_transact::fs::{CopyFrom, Dir, File, Hash, Persist, Restore};
 use tc_transact::{IntoView, Transact, Transaction, TxnId};
 use tc_value::{FloatType, Number, NumberClass, NumberInstance, NumberType};
-use tcgeneric::{TCBoxTryFuture, TCBoxTryStream};
+use tcgeneric::{Instance, TCBoxTryFuture, TCBoxTryStream};
 
 use super::sparse::{DenseToSparse, SparseTensor};
 use super::stream::{Read, ReadValueAt};
 use super::{
     Bounds, Coord, Phantom, Schema, Shape, Tensor, TensorAccess, TensorBoolean, TensorBooleanConst,
     TensorCompare, TensorCompareConst, TensorDualIO, TensorIO, TensorInstance, TensorMath,
-    TensorMathConst, TensorReduce, TensorTransform, TensorType, TensorUnary, ERR_COMPLEX_EXPONENT,
+    TensorMathConst, TensorPersist, TensorReduce, TensorTransform, TensorType, TensorUnary,
+    ERR_COMPLEX_EXPONENT,
 };
 
 use access::*;
 pub use access::{BlockListSparse, DenseAccess, DenseAccessor, DenseWrite};
 pub use file::BlockListFile;
 
-#[allow(unused)]
 mod access;
 mod file;
 mod stream;
 
-/// The number of bytes in one mebibyte.s
+/// The number of bytes in one mebibyte.
 const MEBIBYTE: usize = 1_048_576;
 
 /// The number of elements per dense tensor block, equal to (1 mebibyte / 64 bits).
@@ -91,6 +91,17 @@ where
     }
 }
 
+impl<FD, FS, D, T, B> Instance for DenseTensor<FD, FS, D, T, B>
+where
+    Self: Send + Sync,
+{
+    type Class = TensorType;
+
+    fn class(&self) -> Self::Class {
+        TensorType::Dense
+    }
+}
+
 impl<FD, FS, D, T> DenseTensor<FD, FS, D, T, BlockListFile<FD, FS, D, T>>
 where
     D: Dir,
@@ -134,6 +145,17 @@ where
     }
 }
 
+impl<FD, FS, D, T> TensorPersist for DenseTensor<FD, FS, D, T, DenseAccessor<FD, FS, D, T>> {
+    type Persistent = DenseTensor<FD, FS, D, T, BlockListFile<FD, FS, D, T>>;
+
+    fn as_persistent(self) -> Option<Self::Persistent> {
+        match self.into_inner() {
+            DenseAccessor::File(file) => Some(file.into()),
+            _ => None,
+        }
+    }
+}
+
 impl<FD, FS, D, T, B> TensorAccess for DenseTensor<FD, FS, D, T, B>
 where
     FD: File<Array>,
@@ -774,7 +796,7 @@ where
     type Reduce = DenseTensor<FD, FS, D, T, BlockListReduce<FD, FS, D, T, B>>;
 
     fn product(self, axis: usize) -> TCResult<Self::Reduce> {
-        BlockListReduce::new(self.blocks, axis, DenseTensor::product_all).map(DenseTensor::from)
+        BlockListReduce::product(self.blocks, axis).map(DenseTensor::from)
     }
 
     fn product_all(&self, txn: T) -> TCBoxTryFuture<Number> {
@@ -798,7 +820,7 @@ where
     }
 
     fn sum(self, axis: usize) -> TCResult<Self::Reduce> {
-        BlockListReduce::new(self.blocks, axis, DenseTensor::sum_all).map(DenseTensor::from)
+        BlockListReduce::sum(self.blocks, axis).map(DenseTensor::from)
     }
 
     fn sum_all(&self, txn: T) -> TCBoxTryFuture<Number> {
@@ -1173,6 +1195,8 @@ pub struct DenseTensorView<'en> {
 #[async_trait]
 impl<'en> en::IntoStream<'en> for DenseTensorView<'en> {
     fn into_stream<E: en::Encoder<'en>>(self, encoder: E) -> Result<E::Ok, E::Error> {
+        use en::EncodeSeq;
+
         let mut seq = encoder.encode_seq(Some(2))?;
         seq.encode_element(self.schema)?;
         seq.encode_element(self.blocks)?;
