@@ -5,6 +5,7 @@ use log::debug;
 
 use tc_error::*;
 use tc_transact::fs::Dir;
+use tcgeneric::Tuple;
 
 use super::{TensorAccess, TensorMath, TensorReduce, TensorTransform};
 
@@ -38,22 +39,39 @@ fn parse_format(format: &str) -> TCResult<(Vec<Label>, Label)> {
 
     let f_output = parts.pop_back().unwrap_or("").chars().collect::<Label>();
 
+    let mut present_labels = HashSet::<char>::new();
     let valid_labels: HashSet<char> = VALID_LABELS.iter().cloned().collect();
     for f_input in &f_inputs {
-        let labels: HashSet<char> = f_input.iter().cloned().collect();
-        if labels.len() != f_input.len() {
+        let mut invalid_labels = f_input
+            .iter()
+            .filter(|l| !valid_labels.contains(l))
+            .peekable();
+
+        if invalid_labels.peek().is_some() {
             return Err(TCError::bad_request(
-                "duplicate label in einsum format",
-                f_input.iter().cloned().collect::<String>(),
+                "invalid labels in einsum format",
+                invalid_labels.collect::<Tuple<&char>>(),
             ));
         }
 
-        let invalid_labels = labels.difference(&valid_labels).cloned().collect::<Label>();
-        if !invalid_labels.is_empty() {
-            return Err(TCError::bad_request(
-                "invalid labels in einsum format",
-                invalid_labels.into_iter().collect::<String>(),
-            ));
+        present_labels.extend(f_input);
+    }
+
+    let mut invalid_labels = f_output
+        .iter()
+        .filter(|l| !valid_labels.contains(l))
+        .peekable();
+
+    if invalid_labels.peek().is_some() {
+        return Err(TCError::bad_request(
+            "invalid labels in einsum format",
+            invalid_labels.collect::<Tuple<&char>>(),
+        ));
+    }
+
+    for l in &f_output {
+        if !present_labels.contains(l) {
+            return Err(TCError::bad_request("no such input dimension", l));
         }
     }
 
@@ -90,10 +108,10 @@ fn validate_args<T: TensorAccess>(
         for (label, dim) in f_input.iter().zip(tensor.shape().to_vec().iter()) {
             if let Some(known_dim) = dimensions.get(label) {
                 if *dim != *known_dim {
-                    return Err(TCError::bad_request(
-                        "einsum got inconsistent dimension for axis",
-                        label,
-                    ));
+                    return Err(TCError::unsupported(format!(
+                        "einsum got inconsistent dimension for axis {}: {} vs {}",
+                        label, dim, known_dim
+                    )));
                 }
             } else {
                 dimensions.insert(*label, *dim);
@@ -118,6 +136,7 @@ fn normalize<
         f_input,
         f_output
     );
+
     if f_input == f_output {
         return Ok(tensor);
     }
@@ -202,9 +221,11 @@ where
         assert_eq!(f_input.len(), op.ndim());
 
         if !f_output.contains(&f_input[axis]) {
+            debug!("einsum will contract over axis {}", f_input[axis]);
             op = op.sum(axis)?;
             f_input.remove(axis);
         } else {
+            assert!(f_input.contains(&f_output[axis]));
             axis += 1;
         }
     }
@@ -218,6 +239,7 @@ where
             f_input,
             f_output
         );
+
         let source: HashMap<char, usize> = f_input.iter().cloned().zip(0..f_input.len()).collect();
         let permutation: Vec<usize> = f_output.iter().map(|l| *source.get(l).unwrap()).collect();
         op.transpose(Some(permutation))
