@@ -16,7 +16,7 @@ use tc_error::*;
 use tc_transact::fs::{CopyFrom, Dir, File, Hash, Persist, Restore};
 use tc_transact::{IntoView, Transact, Transaction, TxnId};
 use tc_value::{FloatType, Number, NumberClass, NumberInstance, NumberType};
-use tcgeneric::{TCBoxTryFuture, TCBoxTryStream};
+use tcgeneric::{Instance, TCBoxTryFuture, TCBoxTryStream};
 
 use super::dense::{BlockListSparse, DenseTensor, PER_BLOCK};
 use super::transform;
@@ -26,11 +26,11 @@ use super::{
     TensorMathConst, TensorReduce, TensorTransform, TensorType, TensorUnary, ERR_COMPLEX_EXPONENT,
 };
 
+use crate::TensorPersist;
 use access::*;
 pub use access::{DenseToSparse, SparseAccess, SparseAccessor, SparseWrite};
 pub use table::SparseTable;
 
-#[allow(unused)]
 mod access;
 mod combine;
 mod table;
@@ -55,6 +55,17 @@ impl<FD, FS, D, T, A> SparseTensor<FD, FS, D, T, A> {
     }
 }
 
+impl<FD, FS, D, T, A> Instance for SparseTensor<FD, FS, D, T, A>
+where
+    Self: Send + Sync,
+{
+    type Class = TensorType;
+
+    fn class(&self) -> Self::Class {
+        TensorType::Sparse
+    }
+}
+
 type Condensed<FD, FS, D, T, L, R> =
     DenseTensor<FD, FS, D, T, BlockListSparse<FD, FS, D, T, SparseCombinator<FD, FS, D, T, L, R>>>;
 
@@ -151,6 +162,17 @@ where
     }
 }
 
+impl<FD, FS, D, T> TensorPersist for SparseTensor<FD, FS, D, T, SparseAccessor<FD, FS, D, T>> {
+    type Persistent = SparseTensor<FD, FS, D, T, SparseTable<FD, FS, D, T>>;
+
+    fn as_persistent(self) -> Option<Self::Persistent> {
+        match self.accessor {
+            SparseAccessor::Table(table) => Some(table.into()),
+            _ => None,
+        }
+    }
+}
+
 impl<FD, FS, D, T, A> TensorAccess for SparseTensor<FD, FS, D, T, A>
 where
     FD: File<Array>,
