@@ -1,14 +1,15 @@
 use std::collections::HashMap;
-use std::convert::{TryFrom, TryInto};
+use std::convert::TryInto;
 use std::fmt;
 use std::marker::PhantomData;
 
-use afarray::{Array, CoordBlocks, Coords};
+use afarray::{Array, CoordBlocks, CoordUnique, Coords};
 use async_trait::async_trait;
 use destream::de;
 use futures::future::{self, TryFutureExt};
 use futures::stream::{self, Stream, StreamExt, TryStreamExt};
 use log::debug;
+use safecast::AsType;
 
 use tc_btree::{BTreeType, Node};
 use tc_error::*;
@@ -41,10 +42,11 @@ pub struct SparseTable<FD, FS, D, T> {
 
 impl<FD, FS, D, T> SparseTable<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<BTreeType>,
 {
     /// Create a new `SparseTable` with the given [`Schema`].
@@ -94,10 +96,11 @@ impl<FD, FS, D, T> TensorAccess for SparseTable<FD, FS, D, T> {
 #[async_trait]
 impl<FD, FS, D, T> SparseAccess<FD, FS, D, T> for SparseTable<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
     type Slice = SparseAccessor<FD, FS, D, T>;
@@ -132,12 +135,10 @@ where
         let coords = CoordBlocks::new(coords, shape.len(), PER_BLOCK);
 
         if sort {
-            debug!("SparseTable::filled_at shape is {:?}", shape);
             let coords = sorted_coords::<FD, FS, D, T, _>(&txn, shape.into(), coords).await?;
             Ok(Box::pin(coords))
         } else {
-            debug!("SparseTable::filled_at has no need to sort");
-            Ok(Box::pin(coords))
+            Ok(Box::pin(CoordUnique::new(coords, shape.into(), PER_BLOCK)))
         }
     }
 
@@ -187,10 +188,11 @@ where
 
 impl<FD, FS, D, T> ReadValueAt<D> for SparseTable<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
     type Txn = T;
@@ -231,10 +233,11 @@ where
 #[async_trait]
 impl<FD, FS, D, T, A> CopyFrom<D, SparseTensor<FD, FS, D, T, A>> for SparseTable<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     A: SparseAccess<FD, FS, D, T>,
     D::FileClass: From<BTreeType> + From<TensorType>,
 {
@@ -266,10 +269,11 @@ where
 #[async_trait]
 impl<FD, FS, D, T> Persist<D> for SparseTable<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<BTreeType> + From<TensorType>,
 {
     type Schema = Schema;
@@ -294,10 +298,11 @@ where
 #[async_trait]
 impl<FD, FS, D, T> Restore<D> for SparseTable<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<BTreeType> + From<TensorType>,
 {
     async fn restore(&self, backup: &Self, txn_id: TxnId) -> TCResult<()> {
@@ -308,10 +313,11 @@ where
 #[async_trait]
 impl<FD, FS, D, T> de::FromStream for SparseTable<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<BTreeType> + From<TensorType>,
 {
     type Context = (Self, TxnId);
@@ -376,10 +382,11 @@ impl<FD, FS, D, T> TensorAccess for SparseTableSlice<FD, FS, D, T> {
 #[async_trait]
 impl<FD, FS, D, T> SparseAccess<FD, FS, D, T> for SparseTableSlice<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
     type Slice = SparseAccessor<FD, FS, D, T>;
@@ -447,10 +454,11 @@ where
 
 impl<FD, FS, D, T> ReadValueAt<D> for SparseTableSlice<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
     type Txn = T;
@@ -474,11 +482,12 @@ impl<FD, FS, D, T> fmt::Display for SparseTableSlice<FD, FS, D, T> {
 
 struct SparseTableVisitor<FD, FS, D, T>
 where
-    FD: File<Array>,
-    FS: File<Node> + TryFrom<D::File>,
     D: Dir,
     T: Transaction<D>,
-    D::FileClass: From<BTreeType>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
+    D::FileClass: From<TensorType>,
 {
     table: SparseTable<FD, FS, D, T>,
     txn_id: TxnId,
@@ -487,10 +496,11 @@ where
 #[async_trait]
 impl<FD, FS, D, T> de::Visitor for SparseTableVisitor<FD, FS, D, T>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<BTreeType> + From<TensorType>,
 {
     type Value = SparseTable<FD, FS, D, T>;
@@ -517,13 +527,14 @@ async fn filled_at<'a, FD, FS, D, Txn, T>(
     table: T,
 ) -> TCResult<impl Stream<Item = TCResult<Coord>> + Send + Unpin>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node>,
     D: Dir,
     Txn: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
+    D::FileClass: From<TensorType>,
     T: TableStream,
     T::Selection: TableStream,
-    D::FileClass: From<TensorType>,
 {
     assert!(!axes.is_empty());
     let coords = table.select(axes.into_iter().map(Id::from).collect())?;
