@@ -105,17 +105,12 @@ impl Broadcast {
 
         if source_bounds.iter().all(|bound| bound.is_index()) {
             // can't broadcast a slice with shape []
-            // TODO: can this be handled in the conversion above?
-            source_bounds
-                .into_iter()
-                .map(|bound| match bound {
-                    AxisBounds::At(i) => AxisBounds::In(i..i + 1),
-                    other => other,
-                })
-                .collect()
-        } else {
-            Bounds::from(source_bounds)
+            if let Some(AxisBounds::At(i)) = source_bounds.pop() {
+                source_bounds.push(AxisBounds::In(i..i + 1));
+            }
         }
+
+        Bounds::from(source_bounds)
     }
 
     pub fn invert_coord(&self, coord: &[u64]) -> Coord {
@@ -186,6 +181,26 @@ impl Expand {
             .collect()
     }
 
+    pub fn invert_axis(&self, bounds: &Bounds) -> Option<usize> {
+        debug!("invert expand axis {} in bounds {}", self.expand, bounds);
+        assert!(self.expand < bounds.len());
+
+        if bounds[self.expand].is_index() {
+            return None;
+        }
+
+        let mut expand = self.expand;
+        for bound in &bounds[..self.expand] {
+            if bound.is_index() {
+                expand -= 1;
+            }
+        }
+
+        debug!("bound at expansion index is {:?}", bounds.get(expand));
+
+        Some(expand)
+    }
+
     pub fn shape(&'_ self) -> &'_ Shape {
         &self.shape
     }
@@ -251,14 +266,15 @@ pub struct Reduce {
 impl Reduce {
     pub fn new(source_shape: Shape, axis: usize) -> TCResult<Reduce> {
         if axis >= source_shape.len() {
-            return Err(TCError::bad_request(
-                &format!("tensor with shape {} has no such axis", source_shape),
-                axis,
-            ));
+            return Err(TCError::unsupported(format!(
+                "cannot reduce axis {} of tensor with shape {}",
+                axis, source_shape
+            )));
         }
 
         let mut shape = source_shape.clone();
         shape.remove(axis);
+
         Ok(Reduce {
             source_shape,
             shape,
@@ -276,7 +292,7 @@ impl Reduce {
             .collect()
     }
 
-    pub fn reduce_axis(&self, bounds: &Bounds) -> usize {
+    pub fn invert_axis(&self, bounds: &Bounds) -> usize {
         let elided = bounds[..self.axis]
             .iter()
             .filter(|bound| bound.is_index())
@@ -299,6 +315,10 @@ impl Reduce {
         bounds.insert(self.axis, AxisBounds::all(self.source_shape[self.axis]));
         bounds.into()
     }
+
+    pub fn reduce_axis(&self) -> usize {
+        self.axis
+    }
 }
 
 #[derive(Clone)]
